/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* Copyright (c) 2016 King Kévin <kingkevin@cuvoodoo.info> */

/* standard libraries */
#include <stdint.h> // standard integer types
#include <stdio.h> // standard I/O facilities
#include <stdlib.h> // standard utilities
#include <unistd.h> // standard streams
#include <errno.h> // error number utilities

/* STM32 (including CM3) libraries */
#include <libopencm3/stm32/rcc.h> // real-time control clock library
#include <libopencm3/stm32/gpio.h> // general purpose input output library
#include <libopencm3/cm3/scb.h> // vector table definition
#include <libopencmsis/core_cm3.h> // Cortex M3 utilities
#include <libopencm3/cm3/nvic.h> // interrupt utilities
#include <libopencm3/stm32/exti.h> // external interrupt utilities

/* own libraries */
#include "global.h" // board definitions
#include "usart.h" // USART utilities
#include "usb_cdcacm.h" // USB CDC ACM utilities
#include "led_ws2812b.h" // WS2812b LEDs utilities

/* flag set in interrupts to be processed in main task */
volatile bool button_flag = false; // button has been pressed

// the number of ticks in one second
#define TICKS_PER_SECOND 255
/* I can't do the following in define because how defines are stored would cause an integer overflow in the arithmetics */
const uint32_t ticks_second = TICKS_PER_SECOND; // the number of ticks in one second
const uint32_t ticks_minute = 60*TICKS_PER_SECOND; // the number of ticks in one minute
const uint32_t ticks_hour = 60*60*TICKS_PER_SECOND; // the number of ticks in one hour
const uint32_t ticks_midday = 12*60*60*TICKS_PER_SECOND; // the number of ticks in one midday (12 hours)

uint8_t clock_leds[WS2812B_LEDS*3] = {0}; // RGB values for the WS2812b clock LEDs

/* default output (i.e. for printf) */
int _write(int file, char *ptr, int len)
{
	int i;

	if (file == STDOUT_FILENO || file == STDERR_FILENO) {
		for (i = 0; i < len; i++) {
			if (ptr[i] == '\n') { // add carrier return before line feed. this is recommended for most UART terminals
				usart_putchar_nonblocking('\r'); // a second line feed doesn't break the display
				cdcacm_putchar('\r'); // a second line feed doesn't break the display
			}
			usart_putchar_nonblocking(ptr[i]); // send byte over USART
			cdcacm_putchar(ptr[i]); // send byte over USB
		}
		return i;
	}
	errno = EIO;
	return -1;
}

/* switch on LED */
void led_on(void)
{
#ifdef SYSTEM_BOARD
	gpio_clear(LED_PORT, LED_PIN);
#elif MAPLE_MINI
	gpio_set(LED_PORT, LED_PIN);
#endif
}

/* switch off LED */
void led_off(void)
{
#ifdef SYSTEM_BOARD
	gpio_set(LED_PORT, LED_PIN);
#elif MAPLE_MINI
	gpio_clear(LED_PORT, LED_PIN);
#endif
}

/* toggle LED */
void led_toggle(void)
{
	gpio_toggle(LED_PORT, LED_PIN);
}

/* switch off all clock LEDs */
static void clock_clear(void)
{
	// set all colors of all LEDs to 0
	for (uint16_t i=0; i<LENGTH(clock_leds); i++) {
		clock_leds[i] = 0;
	}
}

/* set hours mark on clock LEDs */
static void clock_hours(void)
{
	for (uint8_t hour=0; hour<12; hour++) {
		uint16_t led = WS2812B_LEDS/12*hour;
		clock_leds[led*3+0] = 0xff;
		clock_leds[led*3+1] = 0xff;
		clock_leds[led*3+2] = 0xff;
	}
}

/* set hour progress on clock LEDs */
static void clock_hour(uint32_t time)
{
	time %= ticks_midday; // the LED strip can only display 12 hours
	uint32_t led_hour = WS2812B_LEDS*time/ticks_midday; // scale to number of LEDs
	if (led_hour<WS2812B_LEDS) {
		clock_leds[led_hour*3+0] = 0;
		clock_leds[led_hour*3+1] = 0;
		clock_leds[led_hour*3+2] = 0xff;
	}
	/*
	for (uint16_t led=0; led<hour; led++) {
		clock_leds[led*3+0] = 0;
		clock_leds[led*3+1] = 0;
		clock_leds[led*3+2] = 0xff;
	}
	*/
}

/* set the LEDs */
static void leds_set(void)
{
	for (uint16_t i=0; i<LENGTH(clock_leds)/3; i++) {
		ws2812b_set_rgb(i,clock_leds[i*3+0],clock_leds[i*3+1],clock_leds[i*3+2]);
	}
}

int main(void)
{	
	SCB_VTOR = (uint32_t) 0x08002000; // relocate vector table because of the bootloader

	rcc_clock_setup_in_hse_8mhz_out_72mhz(); // use 8 MHz high speed external clock to generate 72 MHz internal clock
    usart_setup(); // setup USART (for printing)
    cdcacm_setup(); // setup USB CDC ACM (for printing)
    setbuf(stdout, NULL); // set standard out buffer to NULL to immediately print   
    setbuf(stderr, NULL); // set standard error buffer to NULL to immediately print

	// setup LED
	rcc_periph_clock_enable(LED_RCC); // enable clock for LED
	gpio_set_mode(LED_PORT, GPIO_MODE_OUTPUT_2_MHZ, GPIO_CNF_OUTPUT_PUSHPULL, LED_PIN); // set LED pin to 'output push-pull'
	led_off(); // switch off LED to indicate setup started

	// setup button
#if defined(BUTTON_RCC) && defined(BUTTON_PORT) && defined(BUTTON_PIN) && defined(BUTTON_EXTI) && defined(BUTTON_IRQ)
	rcc_periph_clock_enable(BUTTON_RCC); // enable clock for button
	gpio_set_mode(BUTTON_PORT, GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, BUTTON_PIN); // set button pin to input
	rcc_periph_clock_enable(RCC_AFIO); // enable alternate function clock for external interrupt
	exti_select_source(BUTTON_EXTI, BUTTON_PORT); // mask external interrupt of this pin only for this port
	exti_set_trigger(BUTTON_EXTI, EXTI_TRIGGER_BOTH); // trigger on both edge
	exti_enable_request(BUTTON_EXTI); // enable external interrupt
	nvic_enable_irq(BUTTON_IRQ); // enable interrupt
#endif

	ws2812b_setup(); // setup WS2812b LEDs
	clock_clear(); // clear clock LEDs

	printf("welcome to the CuVoodoo LED clock\n"); // print welcome message
	led_on(); // switch on LED to indicate setup completed

	clock_hour(6*ticks_hour);
	leds_set(); // set the colors of all LEDs
	ws2812b_transmit(); // transmit set color
	
	bool action = false; // if an action has been performed don't go to sleep
	button_flag = false; // reset button flag
	char c; // to store received character
	bool char_flag = false; // a new character has been received
	/* toggle the LED with every transmitted character */
	while (true) { // infinite loop
		while (usart_received) { // echo every received character
			action = true; // action has been performed
			led_toggle(); // toggle LED
			c = usart_getchar(); // store receive character
			char_flag = true; // notify character has been received
		}
		while (cdcacm_received) { // echo every received character
			action = true; // action has been performed
			led_toggle(); // toggle LED
			c = usart_getchar(); // store receive character
			char_flag = true; // notify character has been received
		}
		while (char_flag) {
			char_flag = false; // reset flag
			action = true; // action has been performed
			printf("%c",c); // echo receive character
			switch (c) {
				case 'r': // set red color
					clock_leds[0] = 0xff;
					clock_leds[1] = 0;
					clock_leds[2] = 0;
					break;
				case 'g': // set green color
					clock_leds[0] = 0;
					clock_leds[1] = 0xff;
					clock_leds[2] = 0;
					break;
				case 'b': // set blue color
					clock_leds[0] = 0;
					clock_leds[1] = 0;
					clock_leds[2] = 0xff;
					break;
				case 'm': // show hour marks
					clock_hours();
					break;
				case 'h': // show hours
					clock_hour(6*ticks_hour);
					break;	
				default: // set no color
					clock_clear();
					break;
			}
			leds_set(); // set the colors of all LEDs
			ws2812b_transmit(); // transmit set color
		}
		while (button_flag) {
			button_flag = false; // reset flag
			action = true; // action has been performed
			led_toggle(); // toggle LED
		}
		// go to sleep if nothing had to be done, else recheck for activity
	if (action) {
			action = false;
		} else {
			__WFI(); // go to sleep
		}
	}

	return 0;
}

#if defined(BUTTON_ISR) && defined(BUTTON_EXTI)
void BUTTON_ISR(void)
{
	exti_reset_request(BUTTON_EXTI); // reset interrupt
	button_flag = true; // perform button action
}
#endif
